/*
 Copyright 2007-2008 Foxmarks Inc.

 foxmarks-server.js: component that implements the logical interface to
 the server.

 */

// TO DO:
// * Deal with failure to retrieve baseline state from server -> merge


function Server() {
}

Server.prototype.Cancel = function() {
    if (this.request) {
        this.request.Cancel();
    }
}

Server.baseline = null,     // share the baseline across all instances of server

Server.prototype.Sync = function(callback) {
    var lns = null, sns = null;
    var lcs = null, scs = null;
    var sco = null;
    var lastModified = 0;
    var self = this;
    var fms = Cc["@foxcloud.com/extensions/foxmarks;1"].
        getService(Ci.nsIFoxmarksService);

    // If we've never synced, do an upload or merge.
    if (!gSettings.haveSynced) {
        self.Status(CheckStatus);
    } else {
        // Normal sync: make sure we've got a baseline to work with.
        SetProgressMessage("progress.downloading");
        self._FetchBaseline(GetServerChanges);
        return;
    }

    function CheckStatus(status, response) {
        if (ResponseOK(status, callback)) {
            if (response.isreset == true) {
                self.Upload(callback);
            } else {
                self.Merge(true, callback);
            }
        }
    }

    function GetServerChanges(status) {
        if (ResponseOK(status, callback)) {
            // Get changes.
            self._GetServerChanges(GotServerChanges);
        }
    }

    function GotServerChanges(status, serverContextObject) {
        if (ResponseOK(status, callback)) {
            sco = serverContextObject;
            if (!sco.continuous) {
                // Someone has done an upload. Allow the user the option
                // of performing a merge or a download.

                switch (DiscontinuityPrompt()) {
                case 0: // merge
                    return self.Merge(false, callback);

                case 1: // download
                    return self.Download(callback);

                case 2: // cancel
                    callback(2);
                    return;
                }
            }
            if (!sco.mscs.length && self.prevState == 'ready' && 
                    self._haveFetched) {
                LogWrite("Nothing to see here; move along.");
                gSettings.SyncComplete();
                callback(0);
                return;
            }

            SetProgressMessage("progress.syncing");
            scs = sco.mscs;
            sns = sco.sns;

            // Calculate minimal local change set.
            lastModified = fms.getLastModified();
            lns = new Nodeset();
            lns.FetchFromNative(GotLocalNodeset);
            self._haveFetched = true;
        }
    }

    function GotLocalNodeset(status) {
        if (ResponseOK(status, callback)) {
            var base = new Nodeset(Server.baseline);
            base.Compare(lns, GetLocalCommandset);
        }
    }

    function GetLocalCommandset(status, lcs) {
        if (ResponseOK(status, callback)) {
            // Check to see whether there's been a clobber
            if (lns.length < 3 * Server.baseline.length / 4) {
                LogWrite("Yikes! Length was " + Server.baseline.length + 
                    " but is now " + lns.length);
                self.manual = true;
                if (!ClobberDialog()) {
                    // User canceled. Back off for 24 hours.
                    gBackoffUntil = Date.now() + 24 * 60 * 60 * 1000;
                    callback(2);
                    return;
                }
            }
        
            LogWrite("lcs = " + lcs.length + " scs = " + scs.length);

            try {
                Synchronize(Server.baseline, lcs, scs, lns, sns, 
                    SyncComplete);
            } catch (e) {
                LogWrite("Synchronization failed. Error is " + 
                    e.toSource());
                callback(3);
            }
        }
    }

    function SyncComplete(finalLcs, finalScs, conflicts, showedUI) {
        SetProgressMessage("progress.writing");

        if (showedUI) {
            self.manual = true;
        }

        if (lastModified != fms.getLastModified()) {
            LogWrite("Local datastore changed during sync");
            LogWrite("lastModified " + lastModified + " != " + fms.getLastModified());
            callback(409);
            return;
        }

        // Scoping: assign back to our enclosing function's locals
        lcs = finalLcs;
        scs = finalScs;

        // Apply the server's changes to the local set.
        try {
            lns = new Nodeset(lns);
            scs.execute(lns);
        } catch (e) {
            LogWrite("execute failed: " + e.toSource());
            callback(3);
            return;
        }
        
        // Write local changes (if any) to the server.
        if (lcs.set.length > 0) {
            self._WriteServerChanges(lcs, lns, sco, conflicts, 
                WroteServerChanges);
        } else {
            WroteServerChanges();
        }
    }

    function WroteServerChanges(obj) {
        if (obj) {      // We had something to write
            if (ResponseOK(obj, callback)) {
            } else {
                LogWrite("putchanges failed; response is " + 
                        obj.toJSONString());
                return;
            }
        }

        // Flush the local nodeset back to the native datastore if it changed.
        if (scs.set.length > 0) {
            SetProgressMessage("progress.loading");
            lns.FlushToNative(WroteLocalChanges);
        } else {
            WroteLocalChanges(0);
        }
    }

    function WroteLocalChanges(response) {
        // If we got changes from the server or wrote changes to
        // the server, we'll have a new revision number. If neither
        // of these happened, then the sync was a big no-op.
        if (ResponseOK(response, callback)) {
            if (lcs.set.length > 0 || scs.set.length > 0) {
                self._CompleteTransaction(lns, sco, callback);
            } else {
                // We didn't do anything, but note that we successfully
                // Synced nonetheless.
                gSettings.SyncComplete();
                callback(0);         // Done!
            }
        }
    }

    function DiscontinuityPrompt() {
        var sb = Bundle().GetStringFromName;

        // get a reference to the prompt service component.
        var ps = Cc["@mozilla.org/embedcomp/prompt-service;1"].
          getService(Ci.nsIPromptService);

        var flags = ps.BUTTON_TITLE_IS_STRING * ps.BUTTON_POS_0 +
              ps.BUTTON_TITLE_IS_STRING * ps.BUTTON_POS_1 +
              ps.BUTTON_TITLE_CANCEL    * ps.BUTTON_POS_2;

        rv = ps.confirmEx(null, sb("disc.title"), sb("disc.body"), flags,
            sb("disc.merge"), sb("disc.download"), null, null, {});

        return rv;
    }

    function ClobberDialog() {

        // put up Clobber Dialog and return true if user okays
        // sync, otherwise false.

        var retval = { okay: false };

        var topwin = Cc['@mozilla.org/appshell/window-mediator;1'].
            getService(Ci.nsIWindowMediator).
            getMostRecentWindow(null);
        
        if (!topwin) {
            LogWrite("ClobberDialog: Couldn't find a topwin!");
            return false;
        }


        var win = topwin.openDialog(
            "chrome://foxmarks/content/foxmarks-clobber.xul", "_blank",
            "chrome,dialog,modal,centerscreen", retval);

        LogWrite("ClobberDialog: retval.okay is " + retval.okay);
        
        return retval.okay;
    }
}

function Syncd2Server() {
}
Syncd2Server.prototype = new Server;

function OwnServer() {
}
OwnServer.prototype = new Server;

Syncd2Server.prototype._Args = function(dict) {
    if (this.manual) {
        dict["manual"] = true;
    }
    if (gFailureCount) {
        dict["retry"] = gFailureCount;
    }
    if (gSettings.machineId) {
        if (!dict["log"]) {
            dict["log"] = {};
        }
        dict["log"]["mid"] = gSettings.machineId;
    }
    if (gSettings.viewId) {
        dict.view = gSettings.viewId;
    }
    return dict;
}

Syncd2Server.prototype.Upload = function(callback) {
    var self = this;
    var ns = new Nodeset();
    self.Status(ForceAuth);
    return;

    // Hack: we don't seem to be able to handle
    // authentication during an upload request.
    // So to avoid trouble, we do a throw-away
    // status request first. If auth is required,
    // it will be handled during the status request,
    // after which we will get on to our real work.

    function ForceAuth(status) {
        if (ResponseOK(status, callback)) {
            SetProgressMessage("progress.writing");
            ns.FetchFromNative(Fetched);
            return;
        }
    }

    function Fetched() {
        ns.ProvideCommandset(Continue);
        return;
    }

    function Continue(status, cs) {
        if (ResponseOK(status, callback)) {
            self.request = new Request("POST", 
                { path: "/sync/bookmarks/upload" }, 
                self._Args({ "commands" : cs }) );
            self.request.Start(Done);
        }
    }

    function Done(response) {
        if (ResponseOK(response, callback)) {
            if (!response.toprev) {
                LogWrite("Error: Invalid response to upload");
                callback(500);
            } else {
                self._CompleteTransaction(ns, { toprev: response.toprev }, 
                    callback);
            }
        }
    }
}

Syncd2Server.prototype.Download = function(callback) {
    var self = this;
    var ns;
    var revision;
    SetProgressMessage("progress.downloading");
    self.request = new Request("POST", 
        { path: "/sync/bookmarks/download" }, self._Args({}));
    self.request.Start(Done);
    return true;

    function Done(response) {
        if (ResponseOK(response, callback)) {
            if (!response.toprev || !response.commands) {
                LogWrite("Error: Invalid response to download");
                callback(500);
                return;
            }
            revision = response.toprev;
            ns = new Nodeset();
            var cs = new Commandset(response.commands);
            try {
                cs.execute(ns);
            } catch (e) {
                LogWrite("execute failed: " + e.toSource());
                callback(3);
                return;
            }
            SetProgressMessage("progress.loading");
            ns.FlushToNative(Finished);
        }

        function Finished(response) {
            if (ResponseOK(response, callback)) {
                self._CompleteTransaction(ns, { toprev: revision }, callback);
            }
        }
    }
}


// Fetch changes from the server. Passes to callback an object containing:
// * continous: true if the changes on the server are continuous.
// * mscs: the minimal server commandset. 
// * other server context info (in this case, toprev) to passed into
//   _WriteServerChanges().


Syncd2Server.prototype._GetServerChanges = function(callback) {
    // Get changes.
    var self = this;
    var toprev = 0;
    var sns = new Nodeset(Server.baseline);

    self.request = new Request("POST", 
        { path: "/sync/bookmarks/getchanges" },
        self._Args({ "baserev": Server.baseline.currentRevision}) );
    self.request.Start(ProcessResponse);

    function ProcessResponse(response) {
        if (ResponseOK(response, callback)) {
            if (response.continuous == false) {
                callback(0, { continuous: false });
            } else {
                if (!response.toprev || !response.commands) {
                    LogWrite("Error: Invalid response to getchanges");
                    callback(500);
                    return;
                }
                toprev = response.toprev;
                // Calculate mcs.
                if (response.commands.length) {
                    scs = new Commandset(response.commands);
                    try {
                        scs.execute(sns);
                    } catch (e) {
                        LogWrite("execute failed: " + e.toSource());
                        callback(3);
                        return;
                    }
                    var base = new Nodeset(Server.baseline);
                    base.Compare(sns, CalculatedMcs);
                } else {
                    // Not modified
                    callback(0, { mscs: new Commandset(), toprev: toprev, 
                            sns: sns, continuous: true});
                }
            }
        }
    }

    function CalculatedMcs(status, cs) {
        if (ResponseOK(status, callback)) {
            LogWrite(">>> Finished mcs: " + cs.toSource());
            callback(0, { mscs: cs, toprev: toprev, continuous: true, 
                    sns: sns });
        }
    }
}

OwnServer.prototype._GetServerChanges = function(callback) {
    // Download the file, but only if the etag doesn't match
    var self = this;
    var headers = {};
    var token;
    var etag;
    var sns;

    if (gSettings.etag.length) {
        headers["If-None-Match"] = gSettings.etag;
    }
    self.request = new Request("GET", gSettings.urlWithUsernameAndPassword,
        null, false, headers);
    self.request.Start(ProcessFile);
    return;

    function ProcessFile(response) {
        if (response.status == 304) {   // Not modified.
            callback(0, { continuous: true, mscs: new Commandset(), 
                token: gSettings.token, etag: response.etag });
            return;
        }
        if (ResponseOK(response, callback)) {
            if (response.token != gSettings.token) {
                callback(0, { continuous: false, etag: response.etag } );
                return;
            }

            token = response.token;
            etag = response.etag;
            var cs = new Commandset(response.commands);
            sns = new Nodeset();
            try {
                sns.Execute(cs);
            } catch (e) {
                LogWrite("execute failed: " + e.toSource());
                callback(3);
                return;
            }

            var base = new Nodeset(Server.baseline);
            base.Compare(sns, Compared);
        }
    }

    function Compared(status, cs) {
        if (ResponseOK(status)) {
            callback(0, { continuous: true, mscs: cs, token: token, 
                    sns: sns, etag: etag });
        }
    }
}

Syncd2Server.prototype._WriteServerChanges = function(lcs, lns, sco, conflicts,
        callback) {

    var self = this;
    self.request = new Request("POST", 
        { path: "/sync/bookmarks/putchanges" },
        self._Args({ baserev : sco.toprev, commands : lcs, log : conflicts } ));
    self.request.Start(Wrote);

    function Wrote(response) {
        if (ResponseOK(response, callback)) {
            sco.toprev = response.toprev;
            if (!sco.toprev) {
                LogWrite("Error: Invalid response to putchanges");
                callback(500);
            } else {
                callback(0);
            }
        }
    }
}

OwnServer.prototype._WriteServerChanges = function(lcs, lns, sco, conflicts,
        callback) {

    var self = this;

    lns.ProvideCommandset(GotCommandset);
    return;

    function GotCommandset(status, cs) {
        var header = {};
        if (!gSettings.disableIfMatchOnPut && sco.etag && sco.etag.length) {
            header = { "Etag" : sco.etag };
        }
        if (ResponseOK(status, callback)) {
            self.request = new Request("PUT", 
                gSettings.urlWithUsernameAndPassword,
                { token: sco.token, commands: cs },
                false, header, true);
            self.request.Start(Written);
        }
    }
    function Written(response) {
        if (ResponseOK(response, callback)) {
            sco.etag = response.etag;
            callback(0);
        }
    }
}

Syncd2Server.prototype.Status = function(callback) {
    var self = this;
    LogWrite("Entered Status...");
    SetProgressMessage("progress.verifying");

    self.request = new Request("GET",
        { path: "/sync/bookmarks/status" });
    self.request.Start(CheckStatus);

    function CheckStatus(response) {
        if (ResponseOK(response, callback)) {
            if (!response.toprev) {
                LogWrite("Error: Invalid response to status");
                callback(500);
                return;
            }
            callback(0, response);
        }
    }
}

OwnServer.prototype.Status = function(callback) {
    var self = this;
    LogWrite("Entered status...");
    SetProgressMessage("progress.verifying");

    self.request = new Request("GET",
        gSettings.urlWithUsernameAndPassword);
    self.request.Start(HandleResponse);

    function HandleResponse(response) {
        if (response.status == 404) {
            callback(0, { status: 0, isreset: true });
        } else if (ResponseOK(response, callback)) {
            callback(0, { status: 0, isreset: false });
        }
    }
}


Syncd2Server.prototype._GetServerState = function(callback) {
    var self = this;
    self.request = new Request("POST", 
        { path: "/sync/bookmarks/download" }, self._Args({}) );
    self.request.Start(callback);
}

OwnServer.prototype._GetServerState = function(callback) {
    var self = this;
    self.request = new Request("GET",
        gSettings.urlWithUsernameAndPassword);
    self.request.Start(callback);
}

Server.prototype.Merge = function(local, callback) {
    // Perform a merge (an "additive sync").
    // local is a boolean which determines what merge uses
    // as the starting set: true for local, false for server.

    LogWrite("Entered Merge...");

    var self = this;
    var serverNS = new Nodeset();
    var localNS = new Nodeset();
    var mergedNS = null;
    var revision;
    var sco;
    SetProgressMessage("progress.downloading");
    self._GetServerState(Done);
    return true;

    function Done(response) {
        if (ResponseOK(response, callback)) {
            sco = response;
            var cs = new Commandset(response.commands);
            try {
                cs.execute(serverNS);
            } catch (e) {
                LogWrite("execute failed: " + e.toSource());
                callback(3);
            }
            localNS.FetchFromNative(Fetched);
        }
    }


    function Fetched() {
        SetProgressMessage("progress.merging");
        if (local) {
            localNS.Merge(serverNS);
            mergedNS = localNS;
        } else {
            mergedNS = new Nodeset(serverNS);
            mergedNS.Merge(localNS);
        }

        // calculate server's MCS
        LogWrite("Finished merge; calculating mcs");
        serverNS.Compare(mergedNS, WriteChanges);
    }

    function WriteChanges(status, cs) {
        if (ResponseOK(status, callback)) {
            // Write changes to the server.
            SetProgressMessage("progress.writing");
            if (cs.set.length > 0) {
                self._WriteServerChanges(cs, mergedNS, sco, null, 
                    WroteServerChanges);
            } else {
                WroteServerChanges(0);
            }
        }
    }

    function WroteServerChanges(response) {
        // Write 'em back to native store, too
        if (ResponseOK(response, callback)) {
            SetProgressMessage("progress.loading");
            mergedNS.FlushToNative(Flushed);
        }
    }

    function Flushed(response) {
        if (ResponseOK(response, callback)) {
            self._CompleteTransaction(mergedNS, sco, callback);
        }
    }
}

Syncd2Server.prototype._FetchBaseline = function(callback) {
    // Fetch the baseline if necessary
    var self = this;

    if (Server.baseline && Server.baseline.hash == gSettings.hash) {
        GotIt(0);
        return;
    }

    Server.baseline = new Nodeset();
    try {
        Server.baseline.LoadFromFile();
    } catch (e) {
        // We failed to load our baseline locally -- try getting
        // it from the server.
        LogWrite("Failed to load baseline from file: " + e.name);
        var request = new Request("POST",
            { path: "/sync/bookmarks/download" },
            self._Args({ rev: gSettings.currentRevision, depth: "all", 
                    log: { "error": e.name }}));
        request.Start(LoadedFromServer);
        return;
    }

    Loaded();
    return;

    function LoadedFromServer(response) {
        if (ResponseOK(response, callback)) {
            var cs = new Commandset(response.commands);
            try {
                cs.execute(Server.baseline);
            } catch (e) {
                LogWrite("execute failed: " + e.toSource());
                callback(3);
                return;
            }
            Server.baseline.currentRevision = response.rev;
            Server.baseline.SaveToFile(Loaded);
            return;
        } else {
            LogWrite("Failed to load baseline from server.");
        }
    }

    function Loaded() {
        // Loaded from file or server
        var nat = new NativeDatasource();
        nat.BaselineLoaded(Server.baseline, GotIt);
        return;
    }

    function GotIt(status) {
        if (ResponseOK(status, callback)) {
            Server.baseline.hash = gSettings.hash;
            callback(0);
        }
    }
}

OwnServer.prototype._FetchBaseline = function(callback) {
    // Fetch the baseline if necessary
    var self = this;

    if (Server.baseline && Server.baseline.hash == gSettings.hash) {
        GotIt(0);
        return;
    }

    Server.baseline = new Nodeset();
    try {
        Server.baseline.LoadFromFile();
    } catch (e) {
        // We failed to load our baseline locally -- we're hosed.
        LogWrite("Failed to load baseline.");
        callback(3);
        return;
    }

    Loaded();
    return;

    function Loaded() {
        // Loaded from file 
        var nat = new NativeDatasource();
        nat.BaselineLoaded(Server.baseline, GotIt);
        return;
    }

    function GotIt(status) {
        if (ResponseOK(status, callback)) {
            Server.baseline.hash = gSettings.hash;
            callback(0);
        }
    }
}

Syncd2Server.prototype._CompleteTransaction = function(ns, sco, callback) {
    var self = this;
    ns.Declone(FinishUp);
    return;

    function FinishUp(status) {
        if (ResponseOK(status, callback)) {
            Server.baseline = ns;
            Server.baseline.currentRevision = sco.toprev;
            Server.baseline.SaveToFile(FinishedWrite);
        }
    }

    function FinishedWrite(status) {
        if (ResponseOK(status, callback)) {
            gSettings.currentRevision = sco.toprev;
            gSettings.SyncComplete();
            callback(0);
        }
    }
}

function ResponseOK(response, callback) {
    if (typeof response == 'number') {
        if (!response) {
            return true;
        } else {
            callback(response);
        }
    } else if (typeof response == 'object') {
        if (!response.status) {
            return true;
        } else {
            if (response.status == 403) {
                LogWrite("Got a 403");
                Handle403(response);
            } else if (response.status == 503) {
                try {
                    gServerBackoff = response.backoff_delay || 0;
                    if (gServerBackoff) {
                        LogWrite("Server wants us to wait " +
                            gServerBackoff + " seconds");
                    }
                } catch(e) {}
            }
            callback(response.status);
            return false;
        }
    } else {
        throw Error("unexpected response type: " + response);
    }
}

OwnServer.prototype.Upload = function(callback) {
    var self = this;
    var ns = new Nodeset();
    var token = Date.now().toString(16);
    SetProgressMessage("progress.writing");
    ns.FetchFromNative(Fetched);
    return;

    function Fetched(status) {
        if (ResponseOK(status, callback)) {
            ns.ProvideCommandset(Continue);
            return;
        }
    }

    function Continue(status, cs) {
        if (ResponseOK(status, callback)) {
            self.request = new Request("PUT", 
                gSettings.urlWithUsernameAndPassword,
                { commands: cs, token: token }, false, null, true );
            self.request.Start(Done);
        }
    }

    function Done(response) {
        if (ResponseOK(response, callback)) {
            self._CompleteTransaction(ns, 
                { etag: response.etag, token: token }, callback);
        }
    }
}

OwnServer.prototype.Download = function(callback) {
    var self = this;
    var ns;
    SetProgressMessage("progress.downloading");
    self.request = new Request("GET", gSettings.urlWithUsernameAndPassword);
    self.request.Start(Done);
    return;

    function Done(response) {
        if (ResponseOK(response, callback)) {
            ns = new Nodeset();
            var cs = new Commandset(response.commands);
            try {
                cs.execute(ns);
            } catch (e) {
                LogWrite("execute failed: " + e.toSource());
                callback(3);
                return;
            }
            SetProgressMessage("progress.loading");
            ns.FlushToNative(Finished);
        }

        function Finished(resp) {
            if (ResponseOK(resp, callback)) {
                self._CompleteTransaction(ns, 
                    { etag: response.etag, token: response.token }, callback);
            }
        }
    }
}

OwnServer.prototype._CompleteTransaction = 
        function(ns, sco, callback) {
    var self = this;
    ns.Declone(FinishUp);
    return;

    function FinishUp(status) {
        if (ResponseOK(status, callback)) {
            Server.baseline = ns;
            Server.baseline.SaveToFile(FinishedWrite);
        }
    }

    function FinishedWrite(status) {
        if (ResponseOK(status, callback)) {
            gSettings.token = sco.token;
            if (sco.etag) gSettings.etag = sco.etag;
            gSettings.SyncComplete();
            callback(0);
        }
    }
}


