/* 
 Copyright 2005-2008 Foxmarks Inc.
 
 foxmarks-settings.js: implements FoxmarksSettings, an object that wraps
 access to persistent settings, both user settings and internal stored values.
   
 */

// TO DO:
// * If user changes username or password, delete our cookie.

var Cc = Components.classes;
var Ci = Components.interfaces;
var CCon = Components.Constructor;
const SYNC_REALM = "Foxmarks Sync Login";

function Bundle() {
  var sb = Components.classes["@mozilla.org/intl/stringbundle;1"]
      .getService(Components.interfaces.nsIStringBundleService)
      .createBundle("chrome://foxmarks/locale/foxmarks.properties");
  return sb;
}

// Notify takes the given args object and passes it to observers.
// By convention, args contains at least "status", an integer with
// the following interpretation:
//   0: operation completed successfully.
//   1: operation continues; msg is status update only.
//   2: operation was cancelled by user.
//   other: operation failed.
// Similarly, "msg" contains a user-displayable message.

function Notify(args) {
    var os = Cc["@mozilla.org/observer-service;1"]
        .getService(Ci.nsIObserverService);

    var str = args.toSource();        
    os.notifyObservers(null, "foxmarks-service", str);
}

function SetProgressMessage(msgname) {
    var msg;
    try {
        msg = Bundle().GetStringFromName(msgname);
    } catch(e) {
        msg = "untranslated(" + msgname + ")";
    }

    Notify({status: 1, msg: msg} );
}

function MapError(status) {
    var error = "";

    status = status & 0x0000ffff;

    try {
        error = Bundle().GetStringFromName("error." + status);
    } catch (e) {
        error = Bundle().formatStringFromName("error.unknown", [status], 1);
    }

    return error;
} 

function formatDate(d) {
 /* args to be passed in:
 
 1: Abbreviated month name
 2: Month Number
 3: Day of month
 4: Hours (12 hour clock)
 5: Hours (24 hour clock)
 6: Minutes
 7: am/pm string
 
 Hours24 and minutes are always 2 digits.
 
 So, to get our default en-US time string, we use:
 
 1$S %2$S, %4$S:%6$S %7$S%3$.0S%5$.0S
 
 which gets us something like:
 
 Jan 3, 12:09 pm
 
 */
 
    var rv = "";
    var sb = Bundle();

    var mon = d.getMonth() + 1;
    var day = d.getDate();
    var hours24 = d.getHours();
    var hours;
    var minutes = d.getMinutes();
    
    var ampm = sb.GetStringFromName("date." + (hours24 < 12 ? "am" : "pm"));
    
    if (hours24 > 12)
        hours = hours24 - 12;
    else if (hours24 == 0)
        hours = 12;
    else
        hours = hours24;
    
    if (minutes < 10) minutes = "0" + minutes;
    if (hours24 < 10) hours24 = "0" + hours24;

    var args = [sb.GetStringFromName("date.mon." + mon), String(mon),
        String(day), String(hours), String(hours24), String(minutes), ampm];
    
    rv = sb.formatStringFromName("date.format", args, args.length);
  
    return rv;
}

/**
* Convert a string containing binary values to hex.
* Shamelessly stolen from nsUpdateService.js
*/
function binaryToHex(input) {
    var result = "";
    for (var i = 0; i < input.length; ++i) {
        var hex = input.charCodeAt(i).toString(16);
        if (hex.length == 1)
            hex = "0" + hex;
        result += hex;
    }
    return result;
}

function hex_md5(string) {
    var arr = new Array();
    
    for (var i = 0; i < string.length; ++i)
        arr[i] = string.charCodeAt(i);
        
    try {
        var hash = Components.classes["@mozilla.org/security/hash;1"].
                    createInstance(Components.interfaces.nsICryptoHash);
        hash.initWithString("md5");
        hash.update(arr, arr.length);
        var digest = binaryToHex(hash.finish(false));
    } catch (e) {
        alert(e);
    } 
    return digest.slice(16);    // 8 bytes seems sufficient for our purposes
}

// Return the version string associated with the currently installed version
function FoxmarksExtensionManagerLiteral(value) {
    var rdfs = Components.classes["@mozilla.org/rdf/rdf-service;1"].
        getService(Components.interfaces.nsIRDFService);
    var ds = Components.classes["@mozilla.org/extensions/manager;1"].
        getService(Components.interfaces.nsIExtensionManager).datasource;
    var s = rdfs.GetResource("urn:mozilla:item:foxmarks@kei.com");
    var p = rdfs.GetResource("http://www.mozilla.org/2004/em-rdf#" + value);
    var t = ds.GetTarget(s, p, true);
    if (t instanceof Components.interfaces.nsIRDFLiteral)
        return t.Value;
    else
        return "unknown";
}

function FoxmarksVersion() {
    return FoxmarksExtensionManagerLiteral("version");
}

function FoxmarksUpdateAvailable() {
    var none = ["none", "unknown"];

    return none.indexOf(FoxmarksExtensionManagerLiteral(
            "availableUpdateURL")) < 0;
}

function FoxmarksSettings() {
    var ps = Cc["@mozilla.org/preferences-service;1"].
            getService(Ci.nsIPrefService);
        
    this.prefs = ps.getBranch("foxmarks.");
    this.ps = ps;
}

FoxmarksSettings.prototype = {
    prefs: null,
    
    getCharPref: function(string, def) {
        var result;
        
        try {
            result = this.prefs.getCharPref(string);
        } catch (e) {
            result = def;
        }
        
        return result;
    },

    getIntPref: function(string, def) {
        var result;
        
        try {
            result = this.prefs.getIntPref(string);
        } catch (e) {
            result = def;
        }
        
        return result;
    },


    getBoolPref: function(string, def) {
        var result;
        
        try {
            result = this.prefs.getBoolPref(string);
        } catch (e) {
            result = def;
        }
        
        return result;
    },


    // get fundamental settings

    // Deprecated
    get serverType() {
        return this.getCharPref("serverType", "http");
    },
    
    // Deprecated
    get hostname() {
        return this.getCharPref("hostname", "sync.foxmarks.com");
    },
    
    get username() {
        return this.getCharPref("username", "");
    },


    get sessionPassword() {
        var fms = Cc["@foxcloud.com/extensions/foxmarks;1"].
            getService(Ci.nsIFoxmarksService);

        return fms.getPassword();
    },

    set sessionPassword(pw) {
        var fms = Cc["@foxcloud.com/extensions/foxmarks;1"].
            getService(Ci.nsIFoxmarksService);

        fms.setPassword(pw);
    },

    get password() {

        var pw = this.passwordNoPrompt;

        if (pw != null)
            return pw;

        var username = { value: this.username };
        var password = { value: "" };
        var remember = { value: this.rememberPassword };

        var sb = Bundle().GetStringFromName;
        var rv = Cc["@mozilla.org/embedcomp/prompt-service;1"].
             getService(Ci.nsIPromptService).
             promptUsernameAndPassword(null, sb("appname.long"), 
                 sb("prompt.usernamepassword"),
                 username, password,
                 sb("prompt.rememberpassword"),
                 remember);

        if (!rv) {
            throw Error("User canceled.");
        }

        this.rememberPassword = remember.value;
        this.username = username.value;
        this.password = password.value;

        return password.value;
    },

    get passwordNoPrompt() {
        if (!this.rememberPassword && this.sessionPassword) {
            return this.sessionPassword;
        }

        if (this.rememberPassword) {
            if ("@mozilla.org/passwordmanager;1" in Cc) {
                var host = {};
                var user = {};
                var pass = {}; 
                try {
                    var pmi = Cc["@mozilla.org/passwordmanager;1"].
                        createInstance(Ci.nsIPasswordManagerInternal);
                    pmi.findPasswordEntry(this.host, this.username, 
                        "", host, user, pass);
                    this.sessionPassword = pass.value;
                    return pass.value;
                } catch(e) { }
            } else if ("@mozilla.org/login-manager;1" in Cc) {
                var lm = Cc["@mozilla.org/login-manager;1"].
                    getService(Ci.nsILoginManager);
                var logins = lm.findLogins({}, this.host, null, SYNC_REALM);
                for (var i = 0; i < logins.length; ++i) {
                    if (logins[i].username == this.username) {
                        return logins[i].password;
                    }
                }
            }
        }
        return null;    // couldn't fetch password
    },

    // Deprecated
    get path() {
        return this.getCharPref("path", "/home/{username}/");
    },
    
    get hash() {
        var fms = Cc["@foxcloud.com/extensions/foxmarks;1"].
            getService(Ci.nsIFoxmarksService);

        return hex_md5((this.useOwnServer ? 
                this.url : this.host + this.username) + 
            fms.getStorageEngine())
            + ".";
    },
            
    get lastSynchDate() {
        return this.getCharPref(this.hash + "lastSynchDate", "");
    },

    get haveSynced() {
        return this.lastSynchDate != "";
    },
    
    get minutesSinceLastSync() {
        if (!this.haveSynced)
          return 0;
      
        var syncMS = new Date(this.lastSynchDate).getTime();
        var nowMS = Date.now();
        return (nowMS - syncMS) / 60000;
  
    },
  
    get daysSinceLastUpdateNag() {
        if (!this.lastNagDate)
            return Infinity;
        var updateMS = new Date(this.lastNagDate).getTime();
        var nowMS = Date.now();
        return (nowMS - updateMS) / (60000 * 60 * 24);
    },
  
    get lastNagDate() {
        return this.getCharPref("lastNagDate", null);
    },
    
    set lastNagDate(string) {
        this.prefs.setCharPref("lastNagDate", string);
    },
    
    get lastSynchDisplayDate() {
        if (!this.haveSynced) {
            return Bundle().GetStringFromName("msg.neversynced");
        } else {
            return formatDate(new Date(this.lastSynchDate));
        }
    },

    get etag() {
        return this.getCharPref(this.hash + "etag", "");
    },
    
    get token() {
        return this.getCharPref(this.hash + "token", "");
    },
    
    get writeCount() {
        return this.getIntPref("writeCount", 0);
    },

    get synchOnTimer() {
        return this.getBoolPref("synchOnTimer", true);
    },
    
    get autoSynchFreq() {
        return this.getIntPref("autoSynchFreq", 60);
    },
    
    get syncOnShutdown() {
        return this.getIntPref("syncOnShutdown", true) != 0;
    },

    get syncOnShutdownAsk() {
        return this.getBoolPref("syncOnShutdownAsk", true);
    },
    
    get debugUI() {
        return this.getBoolPref("debugUI", false);
    },
    
    get wizardSuppress() {
        return this.getBoolPref("wizardNoShow", false);
    },
  
    set wizardSuppress(bool) {
       this.prefs.setBoolPref("wizardNoShow", bool);
    },

    get disableIfMatchOnPut() {
        return this.getBoolPref("disableIfMatchOnPut", false);
    },


    get enableLogging() {
        return this.getBoolPref("enableLogging", true);
    },

    get rememberPassword() {
        return this.getBoolPref("rememberPassword", true);
    },

    set rememberPassword(bool) {
        this.prefs.setBoolPref("rememberPassword", bool);
    },

    set username(string) {
        if (string != this.username) {
            this.prefs.setCharPref("username", string);
            this.ClearCredentials();
        }
    },
    
    set password(password) {
        if (this.passwordNoPrompt != password) {
            this.ClearCredentials();
        }
        if (!this.rememberPassword) {
            this.sessionPassword = password;
        } else {
            if ("@mozilla.org/passwordmanager;1" in Cc) {
                var pm = Cc["@mozilla.org/passwordmanager;1"]
                    .createInstance(Ci.nsIPasswordManager);
                try { 
                    pm.removeUser(this.host, this.username);
                } catch(e) {}
                pm.addUser(this.host, this.username, password);
            } else if ("@mozilla.org/login-manager;1" in Cc) {
                // Can't set password without username.
                if (!this.username)
                    return;
                var lm = Cc["@mozilla.org/login-manager;1"].
                    getService(Ci.nsILoginManager);
                var nsli = new CCon("@mozilla.org/login-manager/loginInfo;1",
                    Ci.nsILoginInfo, "init");
                var newli = new nsli(this.host, null, SYNC_REALM, 
                    this.username, password, "", "");
                var oldli = new nsli(this.host, null, SYNC_REALM, 
                    this.username, this.passwordNoPrompt, "", "");
                try {
                    lm.modifyLogin(oldli, newli);
                } catch (e) {
                    lm.addLogin(newli);
                }
            }
        }
    },
    
    ClearCredentials: function() {
        var cm = Cc["@mozilla.org/cookiemanager;1"].
            getService(Ci.nsICookieManager);
        cm.remove(".staging.foxmarks.com", "SYNCD_AUTH", "/", false);
        cm.remove(".foxmarks.com", "SYNCD_AUTH", "/", false);
        return;
    },

    set lastSynchDate(string) {
        this.prefs.setCharPref(this.hash + "lastSynchDate", string);
    },

    set etag(string) {
        this.prefs.setCharPref(this.hash + "etag", string);
    },
    
    set token(string) {
        this.prefs.setCharPref(this.hash + "token", string);
    },

    set writeCount(integer) {
        this.prefs.setIntPref("writeCount", integer);
    },
        
    set autoSynchFreq(integer) {
        this.prefs.setIntPref("autoSynchFreq", integer);
    },
    
    set synchOnTimer(bool) {
        this.prefs.setBoolPref("synchOnTimer", bool);
    },
    
    set syncOnShutdown(integer) {
        this.prefs.setIntPref("syncOnShutdown", integer);
    },
    
    set syncOnShutdownAsk(bool) {
        this.prefs.setBoolPref("syncOnShutdownAsk", bool);
    },

    set debugUI(bool) {
        this.prefs.setBoolPref("debugUI", bool);
    },

    
    // get calculated settings
    get calcPath() {
        return this.path.replace("{username}", this.username);
    },
       
    get NowAsGMT() {
        var d = new Date();
        return d.toGMTString();
    },
    

    SyncComplete: function() {
        this.lastSynchDate = this.NowAsGMT;
    },

    // Additions to support Sync2
    get useOwnServer() {
        return this.getBoolPref("useOwnServer", false);
    },

    set useOwnServer(bool) {
        this.prefs.setBoolPref("useOwnServer", bool);
    },

    set url(u) {
        this.prefs.setCharPref("url", u);
    },

    get url() {
        return this.getCharPref("url", "");
    },

    set hideStatusIcon(b) {
        this.prefs.setBoolPref("hideStatusIcon", b);
        var os = Cc["@mozilla.org/observer-service;1"].
            getService(Ci.nsIObserverService);
        os.notifyObservers(null, "foxmarks-statechange", 
            gSettings.hideStatusIcon ?  "hide" : "show")
    },

    get hideStatusIcon() {
        return this.getBoolPref("hideStatusIcon", false);
    },

    get urlWithUsername() {
        var rep = this.username.length ? this.username + "@" : "";
        return this.url.replace("://", "://" + rep); 
    },

    get urlWithUsernameAndPassword() {
        var user = this.username;
        var pw = this.password;
        
        if (pw.length) {
            user += ":" + pw;
        }

        if (user.length) {
            user += "@";
        }

        return this.url.replace("://", "://" + user);
    },

    get majorVersion() {
        return this.getIntPref("majorVersion", 1);
    },

    set majorVersion(ver) {
        this.prefs.setIntPref("majorVersion", ver);
    },

    get host() {
        if (this.useOwnServer) {
            var exp = /(.*):\/\/([^\/]*)/;
            var result = this.url.match(exp);
            if (!result) {
                return null;
            } else {
                return result[1] + "://" + result[2];
            }
        } else {
            return this.getCharPref("host", "sync.foxmarks.com");
        }
    },

    get acctMgrHost() {
        return this.useOwnServer ? "" : this.getCharPref("acctMgrHost",
            this.host.replace("sync", "login"));
    },

    get wizardUrl() {
        return this.getCharPref("wizardUrl", "https://" + 
            this.acctMgrHost + "/wizard");
    },

    // -1: never synced
    //  0: likely synced in prior installation
    //  > 0: definitely synced
    get currentRevision() {
        return this.getIntPref(this.hash + "currentRevision", -1);
    },

    set currentRevision(cv) {
        if (cv != this.currentRevision) {
            this.prefs.setIntPref(this.hash + "currentRevision", cv);
            this.ps.savePrefFile(null);     // ensure it gets flushed
        }
    },

    get securityLevel() {
        // -1: use cleartext throughout
        //  0: use SSL for auth, cleartext otherwise (default)
        //  1: use SSL everywhere
        return this.getIntPref("securityLevel", 0);
    },

    set securityLevel(level) {
        this.prefs.SetIntPref("securityLevel", level);
    },

    get disableIconSync() {
        return this.getBoolPref("disableIconSync", false);
    },

    get disableDirtyOnBatch() {
        return this.getBoolPref("disableDirtyOnBatch", false);
    },

    get machineId() {
        var id = this.getCharPref("machineId", null);
        if (!id) {
            id = Date.now().toString(36);
            this.prefs.setCharPref("machineId", id);
        }
        return id;
    },

    get viewId() {
        return this.getIntPref(this.hash + "viewId", 0);
    }
}

var gSettings = new FoxmarksSettings();
