/*
 Copyright 2005-2008 Foxmarks Inc.

 foxmarks-service.js: component that implements the "service" interface to
 the core synchronization code.

 */

var Cc = Components.classes;
var Ci = Components.interfaces;

function LoadJavascript(filename, id) {
    if (id == "undefined") {
        Cc["@mozilla.org/moz/jssubscript-loader;1"].
        getService(Ci.mozIJSSubScriptLoader).
        loadSubScript(filename, null);
    }
}

function GetTopWin(wintype) {
    var topwin = Cc['@mozilla.org/appshell/window-mediator;1'].
        getService(Ci.nsIWindowMediator).
        getMostRecentWindow(wintype);

    return topwin;
}

function FoxmarksLaunchSetupWizard() {
    wizardCallback.timer = Cc["@mozilla.org/timer;1"].
        createInstance(Ci.nsITimer);
    wizardCallback.timer.initWithCallback(wizardCallback, 5000,
        Ci.nsITimer.TYPE_ONE_SHOT);
}

var wizardCallback = {
    notify: function(timer) {
        FoxmarksOpenWizard(false, gSettings.username != "");
    }
}

function HandleShutdown(cancel) {
    var retval = { helpurl: null };
    var sb = Bundle().GetStringFromName;
    var dontask = {value: false};
    var rv = 0;

    if (!gSettings.haveSynced || GetState() != "dirty") {
        return;
    }

    var topwin = GetTopWin();

    if (!topwin) {
        LogWrite("HandleShutdown: Couldn't find a topwin!");
        return;
    }

    if (gSettings.syncOnShutdown && gSettings.syncOnShutdownAsk) {

        rv = Cc["@mozilla.org/embedcomp/prompt-service;1"].
        getService(Ci.nsIPromptService).
        confirmEx(topwin, sb("appname.long"), sb("msg.unsynced"),
            Ci.nsIPromptService.STD_YES_NO_BUTTONS, null, null, null,
            sb("msg.dontask"), dontask);
        // Reverse sense: confirmEx returns 0 - yes, 1 - no
        rv = !rv;

        // If user says "don't ask me again", set syncOnShutdown to whatever
        // they have chosen in this instance.
        if (dontask.value) {
            gSettings.syncOnShutdown = rv;
        }
        gSettings.syncOnShutdownAsk = !dontask.value;

    } else {                           // don't ask
        rv = gSettings.syncOnShutdown;
    }

    if (rv) {
        var win = topwin.openDialog(
            "chrome://foxmarks/content/foxmarks-progress.xul", "_blank",
            "chrome,dialog,modal,centerscreen", "synch", retval, null);
        if (retval.helpurl) { // we hit an error and user pressed help button
            if (cancel instanceof Ci.nsISupportsPRBoolean) {
                cancel.value = true;
                topwin.openDialog("chrome://browser/content/browser.xul",
                    "_blank", "chrome,all,dialog=no", retval.helpurl);
            }
        }
    }
}

function LoadFiles() {
    LoadJavascript("chrome://foxmarks/content/foxmarks-log.js",
        typeof(LogWrite));
    LoadJavascript("chrome://foxmarks/content/foxmarks-settings.js",
        typeof(gSettings));
    LoadJavascript("chrome://foxmarks/content/foxmarks-update.js",
        typeof(ForceUpdate));
    LoadJavascript("chrome://foxmarks/content/foxmarks-clobber.js",
        typeof(onClobberCancel));
    if ("@mozilla.org/browser/nav-bookmarks-service;1" in Cc)
        LoadJavascript("chrome://foxmarks/content/foxmarks-places.js",
            typeof(NativeDatasource));
    else
        LoadJavascript("chrome://foxmarks/content/foxmarks-rdf.js",
            typeof(NativeDatasource));
    LoadJavascript("chrome://foxmarks/content/foxmarks-nodes.js",
        typeof(Node));
    LoadJavascript("chrome://foxmarks/content/foxmarks-command.js",
        typeof(Command));
    LoadJavascript("chrome://foxmarks/content/foxmarks-core.js",
        typeof(Synchronize));
    LoadJavascript("chrome://foxmarks/content/foxmarks-network.js",
        typeof(Request));
    LoadJavascript("chrome://foxmarks/content/foxmarks-server.js",
        typeof(Server));
    LoadJavascript("chrome://foxmarks/content/foxmarks-json.js",
        "undefined");
    LoadJavascript("chrome://foxmarks/content/foxmarks-utils.js",
        typeof(forEach));
    LoadJavascript("chrome://foxmarks/content/foxmarks-uitools.js",
        typeof(FoxmarksOpenWindowByType));
}

var logStream = null;

function logFileOpen() {
    var file = Cc['@mozilla.org/file/directory_service;1']
    .getService(Ci.nsIProperties)
    .get('ProfD', Ci.nsIFile);

    file.append("foxmarks.log");

    logStream = Cc["@mozilla.org/network/file-output-stream;1"]
    .createInstance(Ci.nsIFileOutputStream);

    // use write, append, create
    logStream.init(file, 0x02 | 0x08 | 0x10, 0664, 0);
}

function logFileClose() {
    logStream.close();
    logStream = null;
}

var gFailureCount = 0;// number of consecutive times we've failed
var gBackoffUntil = 0;// if we've been failing, our backoff time (ms since 1970)
var gServerBackoff = 0;  // seconds server wants us to back off

function ReturnErrorMsg(code, msg) {
    Notify({status: code, msg: msg });
    ClearBusy();
    SetState(code == 503 ? "dirty" : "error");
    LogWrite("Returned error: " + msg + "(" + code + ")");
    var d = new Date();
    // Initial back-off is 15 minutes, doubling with each error
    gBackoffUntil = d.getTime() + 1000 * (Math.max(
        15 * 60 * Math.pow(2, gFailureCount++),
        gServerBackoff) + Math.floor(Math.random() * 15));
    gServerBackoff = 0;
    var retry = new Date();
    retry.setTime(gBackoffUntil);
    LogWrite("Will retry at " + retry);
    fms.lastError = code;
}

function ReturnErrorCode(code) {
    ReturnErrorMsg(code, MapError(code));
}

function ReturnSuccess(msgname, args) {
    if (args == null) {
        var args = {};
    }

    args.status = 0;
    args.msg = Bundle().GetStringFromName(msgname);

    Notify(args);
    SetState("ready");
    ClearBusy();
    gFailureCount = 0;
    fms.lastError = 0;
    LogWrite("Success: " + Bundle().GetStringFromName(msgname));
}

function SetBusy() {
    if (IsBusy._busy) {
        return false;
    } else {
        IsBusy._busy = true;
        SetState("working");
        return true;
    }
}

function ClearBusy() {
    IsBusy._busy = false;
}

function IsBusy() {
    return IsBusy._busy;
}
IsBusy._busy = false;

function GetState() {
    return GetState._state;
}
GetState._state = "ready";

function SetState(newstate) {
    if (newstate == GetState()) {
        return;
    }

    var os = Cc["@mozilla.org/observer-service;1"]
    .getService(Ci.nsIObserverService);

    os.notifyObservers(null, "foxmarks-statechange", newstate);

    GetState._state = newstate;
}

// Internal callbacks

function LogStart(op, dest) {
    LogWrite("------ Foxmarks/" + FoxmarksVersion() + " (" + 
            fms.getStorageEngine() + ") starting " + op +
            " with " + dest + " ------");
}

///////////////////////////////////////////////////////////////////////////
//
// nsFoxmarksService
//

fms = null;  // set during initialization to instance object

function nsFoxmarksService() {
    LoadFiles();
}

nsFoxmarksService.prototype = {
    /////////////////////////////////////////////////////////////////////////
    // nsIFoxmarksService

    timer: null,
    _server: null,
    lastmodified: null,

    status: function() {
        // return if we're currently processing another request
        if (!SetBusy()) {
            return false;
        }
        LogStart("status", gSettings.host);
        this.server.Status(Finished);
        return true;

        function Finished(status, response) {
            if (status) {
                ReturnErrorCode(status);
            } else {
                ReturnSuccess("msg.accountverified", response);
            }
        }
    },

    synchronize: function(automatic) {
        var prevState = GetState();

        // return if we're currently processing another request
        if (!SetBusy()) {
            return false;
        }

        LogStart("sync", gSettings.host);
        this.server.manual = automatic === true ? false : true;
        this.server.prevState = prevState;
        this.server.Sync(Finished);
        return true;

        function Finished(status) {
            if (!status) {
                ReturnSuccess("msg.synccompleted");
            } else {
                ReturnErrorCode(status);
            }
        }
    },

    synchronizeInitial: function (remoteIsMaster, doMerge) {
        // return if we're currently processing another request
        if (!SetBusy()) {
            return false;
        }

        LogStart("initial sync", gSettings.host);
        this.server.manual = true;

        if (doMerge) {
            this.server.Merge(!remoteIsMaster, Finished);
        } else {
            if (remoteIsMaster) {
                this.server.Download(Finished);
            } else {
                this.server.Upload(Finished);
            }
        }

        return true;

        function Finished(status) {
            if (!status) {
                ReturnSuccess("msg.synccompleted");
            } else {
                ReturnErrorCode(status);
            }
        }
    },

    upload: function () {
        // return if we're currently processing another request
        if (!SetBusy()) {
            return false;
        }

        LogStart("upload", gSettings.host);

        this.server.manual = true;
        this.server.Upload(Finished);
        return true;

        function Finished(status) {
            if (!status) {
                ReturnSuccess("msg.uploadcompleted");
            } else {
                ReturnErrorCode(status);
            }
        }
    },

    download: function () {
        // return if we're currently processing another request
        if (!SetBusy()) {
            return false;
        }

        LogStart("download", gSettings.host);

        this.server.manual = true;
        this.server.Download(Finished);
        return true;

        function Finished(status) {
            if (!status) {
                ReturnSuccess("msg.remotefilecopied");
            } else {
                ReturnErrorCode(status);
            }
        }
    },

    cancel: function() {
        this.server.Cancel();
    },

    logWrite: function (msg) {
        if (!logStream)
            logFileOpen();

        var d = new Date();
        var year = d.getFullYear();
        var month = d.getMonth() + 1; if (month < 10) month = "0" + month;
        var day = d.getDate(); if (day < 10) day = "0" + day;
        var hour = d.getHours(); if (hour < 10) hour = "0" + hour;
        var minute = d.getMinutes(); if (minute < 10) minute = "0" + minute;
        var sec = d.getSeconds(); if (sec < 10) sec = "0" + sec;

        // format is [YYYY-MM-DD HH:MM:SS] msg\n
        var string = "[" + year + "-" + month + "-" + day + " " + hour + ":" +
            minute + ":" + sec + "] " + msg + "\n";

        logStream.write(string, string.length)
    },

    getState: function() {
        return GetState();
    },

    _password: null,

    getPassword: function() {
        return this._password;
    },

    setPassword: function(password) {
        this._password = password;
    },

    getStorageEngine: function() {
        return NativeDatasource.STORAGE_ENGINE;
    },

    getLastError: function() {
        return this.lastError;
    },

    lastError: 0,

    getLastModified: function() {
        return fms.lastmodified;
    },

    get server() {
        var cls = gSettings.useOwnServer ? OwnServer : Syncd2Server;
        if (!(this._server instanceof cls)) {
            this._server = new cls();
        }
        return this._server;
    },

    /////////////////////////////////////////////////////////////////////////
    //
    // nsIObserver

    observe: function(subject, topic, data)  { // called at startup
        var timerCallback = {

            notify: function(timer) {
                var now = new Date().getTime();

                // scan entire bookmark set to find
                // last modified date for entire set
                // XXX: Implement

                // Do update nag if necessary
                if (gSettings.daysSinceLastUpdateNag > 28 &&
                    FoxmarksUpdateAvailable()) {
                    gSettings.lastNagDate = gSettings.NowAsGMT;
                    UpdateNag();
                }

                // Do automatic sync if necessary
                if (!IsBusy() && (!gFailureCount || now > gBackoffUntil) &&
                    gSettings.synchOnTimer && gSettings.haveSynced) {
                    if (gSettings.minutesSinceLastSync > 
                            gSettings.autoSynchFreq) {
                        fms.synchronize(true);
                    } else {
                        if (fms.lastmodified && gSettings.haveSynced && 
                            fms.lastmodified >
                            Date.parse(gSettings.lastSynchDate) &&
                            now - fms.lastmodified > 5 * 60 * 1000) {
                            fms.synchronize(true);
                        }
                    }
                }
            }
        }

        if (topic == "app-startup") {
            // Pre-initialization here.
            var os = Cc["@mozilla.org/observer-service;1"].
                getService(Ci.nsIObserverService);
            os.addObserver(this, "quit-application-requested", false);
            os.addObserver(this, "foxmarks-datasourcechanged", false);
            os.addObserver(this, "final-ui-startup", false);
        } else if (topic == "final-ui-startup") {
            // Real initialization starts here.
            fms = this;
            fms.nat = new NativeDatasource();
            fms.nat.WatchForChanges();

            if (!gSettings.wizardSuppress && !gSettings.useOwnServer &&
                    !gSettings.haveSynced) {
                FoxmarksLaunchSetupWizard();
            } else if (gSettings.majorVersion < 2) {
                gSettings.majorVersion = 2;
                if (gSettings.hostname == "sync.foxmarks.com" || 
                        gSettings.hostname == "sync.foxcloud.com")  {
                    // Standard upgrade
                    FoxmarksLaunchSetupWizard();
                } else {
                    // Custom server upgrade
                    var sb = Bundle();
                    Cc["@mozilla.org/embedcomp/prompt-service;1"].
                        getService(Ci.nsIPromptService).
                        alert(null, sb("appname.long"), 
                        sb("msg.upgrade2custom"));
                    gSettings.useOwnServer = true;
                    gSettings.url = gSettings.serverType + "://" + 
                        gSettings.hostname + 
                        gSettings.path.replace("{username}", 
                            gSettings.username) +
                        "foxmarks.json";
                }
            }

            this.timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
            this.timer.initWithCallback(timerCallback, 1000*60,
                Ci.nsITimer.TYPE_REPEATING_SLACK);
        } else if (topic == "quit-application-requested") {
            HandleShutdown();
        } else if (topic == "foxmarks-datasourcechanged") {
            this.lastmodified = parseInt(data);
            var okState = (GetState() == "ready" || GetState() == "unknown");
            if (okState && gSettings.haveSynced &&
                this.lastmodified > Date.parse(gSettings.lastSynchDate)) {
                SetState("dirty");
            }
        } else {
            LogWrite("Yikes unknown topic " + topic);
        }
    },


    /////////////////////////////////////////////////////////////////////////
    // nsIClassInfo
    getInterfaces: function (aCount) {
        var interfaces = [Ci.nsIFoxmarksService,
        Ci.nsIObserver,
        Ci.nsiRDFObserver];
        aCount.value = interfaces.length;
        return interfaces;
    },

    getHelperForLanguage: function (aLanguage) {
        return null;
    },

    get contractID() {
        return "@foxcloud.com/extensions/foxmarks;1";
    },

    get classDescription() {
        return "Foxmarks Service";
    },

    get classID() {
        return Components.ID("{49ace257-6111-48b2-a988-f9eb38b0fa58}");
    },

    get implementationLanguage() {
        return Ci.nsIProgrammingLanguage.JAVASCRIPT;
    },

    get flags() {
        return Ci.nsIClassInfo.SINGLETON;
    },

    /////////////////////////////////////////////////////////////////////////
    // nsISupports
    QueryInterface: function (aIID) {
        if (!aIID.equals(Ci.nsIFoxmarksService) &&
            !aIID.equals(Ci.nsISupports) &&
            !aIID.equals(Ci.nsIRDFObserver) &&
            !aIID.equals(Ci.nsIObserver))
        throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
};

var gModule = {
    _firstTime: true,

    registerSelf: function (aComponentManager, aFileSpec, aLocation, aType) {
        if (this._firstTime) {
            this._firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        }


        aComponentManager = aComponentManager.
        QueryInterface(Ci.nsIComponentRegistrar);

        for (var key in this._objects) {
            var obj = this._objects[key];
            aComponentManager.registerFactoryLocation(obj.CID, obj.className,
                obj.contractID, aFileSpec, aLocation, aType);
        }

        // Make the Foxmarks Service a startup observer
        var cm = Cc["@mozilla.org/categorymanager;1"].
        getService(Ci.nsICategoryManager);
        cm.addCategoryEntry("app-startup", this._objects.service.className,
            "service," + this._objects.service.contractID, true, true, null);
    },


    getClassObject: function (aComponentManager, aCID, aIID) {
        if (!aIID.equals(Ci.nsIFactory))
            throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

        for (var key in this._objects) {
            if (aCID.equals(this._objects[key].CID))
                return this._objects[key].factory;
        }

        throw Components.results.NS_ERROR_NO_INTERFACE;
    },

    _makeFactory: #1= function(ctor) {
        return {
            createInstance: function (outer, iid) {
                if (outer != null)
                    throw Components.results.NS_ERROR_NO_AGGREGATION;
                return (new ctor()).QueryInterface(iid);
            }
        };
    },

    _objects: {
        service: { CID : nsFoxmarksService.prototype.classID,
            contractID : nsFoxmarksService.prototype.contractID,
            className  : nsFoxmarksService.prototype.classDescription,
            factory    : #1#(nsFoxmarksService)
        }
    },

    canUnload: function (aComponentManager)
    {
        return true;
    }
};

function NSGetModule(compMgr, fileSpec)
{
    return gModule;
}
